#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <errno.h>

#include "newinfi.h"

int
read_request (int csock, char buf[], size_t buf_size, size_t *buf_len)
     /* Read from csock and into buf until either buf_size characters
      * have been read, or MAXDELAY seconds have elapsed, or two \n
      * characters have been read in sequence (possibly with \r's
      * between them), or the socket has been closed, or an error
      * happens.  Store the number of characters read in *buf_len if
      * buf_len is not NULL.  Return 1 in case of timeout, 2 if
      * buf_size characters have been read witout encountering two
      * \n's, 3 if the socket was closed under our nose, -1 in case of
      * error (in which case errno is positioned to indicate the
      * cause), and 0 in case the two \n's were received
      * successfully. */
{
  struct timeval start, stop, now;
  size_t len;
  struct pollfd wc;
  int pv;
  ssize_t v;
  size_t i;
  char ln;

  if ( gettimeofday (&start, NULL) < 0 )
    {
      perror ("gettimeofday");
      exit (EXIT_FAILURE);
    }
  stop.tv_sec = start.tv_sec + MAXDELAY;
  stop.tv_usec = start.tv_usec;
  len = 0;
  while ( 1 )
    {
      if ( gettimeofday (&now, NULL) < 0 )
	{
	  perror ("gettimeofday");
	  exit (EXIT_FAILURE);
	}
      if ( now.tv_sec > stop.tv_sec
	   || ( now.tv_sec == stop.tv_sec && now.tv_usec > stop.tv_usec ) )
	{
	  if ( buf_len )
	    *buf_len = len;
	  return 1;
	}
      wc.fd = csock;
      wc.events = POLLIN;
      pv = poll (&wc, 1, ((stop.tv_sec-now.tv_sec)*1000000
			  + (stop.tv_usec-now.tv_usec))/1000);
      if ( pv < 0 )
	{
	  if ( errno == EINTR || errno == ERESTART )
	    continue;
	  perror ("poll");
	  if ( buf_len )
	    *buf_len = len;
	  return -1;
	}
      else if ( pv == 0 )
	{
	  if ( buf_len )
	    *buf_len = len;
	  return 1;
	}
      v = read (csock, buf+len, buf_size-len);
      if ( v < 0 )
	{
	  if ( errno == EINTR || errno == ERESTART )
	    continue;
	  perror ("read");
	  if ( buf_len )
	    *buf_len = len;
	  return -1;
	}
      else if ( v == 0 )
	{
	  if ( buf_len )
	    *buf_len = len;
	  return 3;
	}
      else
	len += v;
      ln = 1;
      for ( i=0 ; i<len ; i++ )
	{
	  if ( buf[i] == '\n' )
	    {
	      if ( ln )
		{
		  if ( buf_len )
		    *buf_len = i;
		  return 0;
		}
	      else
		ln = 1;
	    }
	  else if ( buf[i] != '\r' )
	    ln = 0;
	}
      if ( len >= buf_size )
	{
	  if ( buf_len )
	    *buf_len = buf_size;
	  return 2;
	}
    }
}

int
write_response (int csock, const char buf[], size_t buf_len, size_t *buf_wlen)
     /* Write from buf and into csock until buf_len characters have
      * been read, or MAXDELAY seconds have elapsed, or the socket has
      * been closed, or an error happens.  Store the number of
      * characters written in *buf_wlen if buf_wlen is not NULL.
      * Return 1 in case of timeout, 3 if the socket was closed under
      * our nose, -1 in case of error (in which case errno is
      * positioned to indicate the cause), and 0 in case all buf_len
      * characters have been successfully written. */
{
  struct timeval start, stop, now;
  size_t wlen;
  struct pollfd wc;
  int pv;
  ssize_t v;

  if ( gettimeofday (&start, NULL) < 0 )
    {
      perror ("gettimeofday");
      exit (EXIT_FAILURE);
    }
  stop.tv_sec = start.tv_sec + MAXDELAY;
  stop.tv_usec = start.tv_usec;
  wlen = 0;
  while ( 1 )
    {
      if ( gettimeofday (&now, NULL) < 0 )
	{
	  perror ("gettimeofday");
	  exit (EXIT_FAILURE);
	}
      if ( now.tv_sec > stop.tv_sec
	   || ( now.tv_sec == stop.tv_sec && now.tv_usec > stop.tv_usec ) )
	{
	  if ( buf_wlen )
	    *buf_wlen = wlen;
	  return 1;
	}
      wc.fd = csock;
      wc.events = POLLOUT;
      pv = poll (&wc, 1, ((stop.tv_sec-now.tv_sec)*1000000
			  + (stop.tv_usec-now.tv_usec))/1000);
      if ( pv < 0 )
	{
	  if ( errno == EINTR || errno == ERESTART )
	    continue;
	  perror ("poll");
	  if ( buf_wlen )
	    *buf_wlen = wlen;
	  return -1;
	}
      else if ( pv == 0 )
	{
	  if ( buf_wlen )
	    *buf_wlen = wlen;
	  return 1;
	}
      v = write (csock, buf+wlen, buf_len-wlen);
      if ( v < 0 )
	{
	  if ( errno == EINTR || errno == ERESTART )
	    continue;
	  perror ("write");
	  if ( buf_wlen )
	    *buf_wlen = wlen;
	  return -1;
	}
      else if ( v == 0 )
	{
	  if ( buf_wlen )
	    *buf_wlen = wlen;
	  return 3;
	}
      else
	wlen += v;
      if ( wlen >= buf_len )
	{
	  if ( buf_wlen )
	    *buf_wlen = buf_len;
	  return 0;
	}
    }
}
